/*
    $(CC) -I/usr/include/libxml2 -I. -L ../../lib -l xml2 tv_scan.c dvb.c xml.c -o tv_scan


*/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>

#include "dvb.h"
#include "parse_dvb.h"
//#include "dump_dvb.h"

//----------------------------------------------------------------------------
EnumType inversion2_str[] = {
    { 3,              "reserved",  },
    { INVERSION_OFF,  "INVERSION_OFF",  },
    { INVERSION_ON,   "INVERSION_ON",   },
    { INVERSION_AUTO, "INVERSION_AUTO", }
};

EnumType bandwidth2_str[] = {
    { 3,               "reserved",  },
    { BANDWIDTH_6_MHZ, "BANDWIDTH_6_MHZ", },
    { BANDWIDTH_7_MHZ, "BANDWIDTH_7_MHZ", },
    { BANDWIDTH_8_MHZ, "BANDWIDTH_8_MHZ", }
};

EnumType code_rate2_str[] = {
    { 10,       "reserved",  },
    { FEC_1_2,  "FEC_1_2",  },
    { FEC_2_3,  "FEC_2_3",  },
    { FEC_3_4,  "FEC_3_4",  },
    { FEC_4_5,  "FEC_4_5",  },
    { FEC_5_6,  "FEC_5_6",  },
    { FEC_6_7,  "FEC_6_7",  },
    { FEC_7_8,  "FEC_7_8",  },
    { FEC_8_9,  "FEC_8_9",  },
    { FEC_AUTO, "FEC_AUTO", },
    { FEC_NONE, "FEC_NONE", }
};

EnumType guard_interval2_str[] = {
    { 5,                   "reserved",  },
    { GUARD_INTERVAL_1_16, "GUARD_INTERVAL_1_16", },
    { GUARD_INTERVAL_1_32, "GUARD_INTERVAL_1_32", },
    { GUARD_INTERVAL_1_4,  "GUARD_INTERVAL_1_4",  },
    { GUARD_INTERVAL_1_8,  "GUARD_INTERVAL_1_8",  },
    { GUARD_INTERVAL_AUTO, "GUARD_INTERVAL_AUTO", }
};

EnumType hierarchy2_str[] = {
    { 5,              "reserved",  },
    { HIERARCHY_1,    "HIERARCHY_1",    },
    { HIERARCHY_2,    "HIERARCHY_2",    },
    { HIERARCHY_4,    "HIERARCHY_4",    },
    { HIERARCHY_NONE, "HIERARCHY_NONE", },
    { HIERARCHY_AUTO, "HIERARCHY_AUTO", }
};

EnumType constellation2_str[] = {
    { 7,        "reserved",  },
    { QPSK,     "QPSK",       },
    { QAM_128,  "QAM_128",    },
    { QAM_16,   "QAM_16",     },
    { QAM_256,  "QAM_256",    },
    { QAM_32,   "QAM_32",     },
    { QAM_64,   "QAM_64",     },
    { QAM_AUTO, "QAM_AUTO",   }
};

EnumType transmission_mode2_str[] = {
    { 3,                     "reserved",  },
    { TRANSMISSION_MODE_2K,   "TRANSMISSION_MODE_2K",    },
    { TRANSMISSION_MODE_8K,   "TRANSMISSION_MODE_8K",    },
    { TRANSMISSION_MODE_AUTO, "TRANSMISSION_MODE_AUTO",  }
};
//----------------------------------------------------------------------------

unsigned char **get_section(char *dmxdev, char *extstr, int pid, int tid, int mask, int *pnum_sections)
{
    int fd, count;
    int section_read = 0;
    int section_length;
    int section_number;
    int last_section_number;
    unsigned char *buf;
    unsigned char **buf_list;
    struct dmx_sct_filter_params f;
    int first_section = 1;
    int num_sections;
    int got_all_sections = 0;
    int i;


    memset(&f, 0, sizeof(f));
    f.pid = pid;
    f.filter.filter[0] = tid;
    f.filter.mask[0] = mask;
    f.timeout = 0;
    f.flags = DMX_IMMEDIATE_START | DMX_CHECK_CRC;

    if ((fd = open(dmxdev, O_RDWR)) < 0) {
        printf("\nopening pat demux failed\n");
        return NULL;
    }
    if (ioctl(fd, DMX_SET_FILTER, &f) == -1) {
        printf("\nioctl DMX_SET_FILTER failed\n");
        close(fd);
        return NULL;
    }

    do {
        buf = malloc(4096);
        if (!buf) {
            printf("\nmalloc failed\n");
            return NULL;
        }

        section_read = 0;
        do {
            if (((count = read(fd, buf, 4096)) < 0) && errno == EOVERFLOW) {
                count = read(fd, buf, 4096);
            }
            if (count < 0) {
                printf("\nread_sections: read error\n");
                close(fd);
                return NULL;
            }
            section_length = 3 + (((buf[1] & 0x0f) << 8) | buf[2]);
            section_number = buf[6];
            last_section_number = buf[7];

            if (count == section_length) {
                section_read = 1;
            }
            else {
                printf(".");
            }
        } while (!section_read && !timeout_flag);
        //if 1st section
        //  create a list of sections
        if (first_section) {
            first_section = 0;
            num_sections = 1+last_section_number;
            buf_list = malloc(num_sections * sizeof(unsigned char *));
            memset(buf_list, 0, num_sections * sizeof(unsigned char *));
        }
        buf_list[section_number] = buf;
        got_all_sections = 1;
        for (i=0; i<num_sections; i++) {
            if (buf_list[i] == 0) got_all_sections = 0;
        }
        if (!got_all_sections) printf(".");
    } while(!got_all_sections && !timeout_flag);

    *pnum_sections = num_sections;
    close(fd);
    return buf_list;
}


int set_pesfilter(int fd, int pid, dmx_pes_type_t type, int dvr)
{
        struct dmx_pes_filter_params pesfilter;

//        if (pid <= 0 || pid >= 0x1fff)
//                return 0;

        pesfilter.pid = pid;
        pesfilter.input = DMX_IN_FRONTEND;
        pesfilter.output = dvr ? DMX_OUT_TS_TAP : DMX_OUT_DECODER;
        pesfilter.pes_type = type;
        pesfilter.flags = DMX_IMMEDIATE_START;
        if (ioctl(fd, DMX_SET_PES_FILTER, &pesfilter) < 0) {
            printf("ioctl(DMX_SET_PES_FILTER) for %s PID failed\n",
                type == DMX_PES_AUDIO ? "Audio" :
                type == DMX_PES_VIDEO ? "Video" :
                type == DMX_PES_OTHER ? "Other" : "??");
            return -1;
        }

        return 0;
}


int setup_frontend(int fe_fd, struct dvb_frontend_parameters *frontend)
{
    struct dvb_frontend_info fe_info;

    if (ioctl(fe_fd, FE_GET_INFO, &fe_info) < 0) {
        printf("\nsetup_frontend(), ioctl FE_GET_INFO failed, fe_fd=%x, errno=%d\n", fe_fd, errno);
        return -1;
    }

    if (fe_info.type != FE_OFDM) {
        printf("\nsetup_frontend(), frontend device is not a OFDM (DVB-T) device\n");
        return -1;
    }

    if (ioctl(fe_fd, FE_SET_FRONTEND, frontend) < 0) {
        printf("\nsetup_frontend(), ioctl FE_SET_FRONTEND failed\n");
        return -1;
    }

    return 0;
}


int print_frontend_stats(int fe_fd)
{
    fe_status_t status;
    uint16_t snr, _signal;
    uint32_t ber, uncorrected_blocks;

    ioctl(fe_fd, FE_READ_STATUS, &status);
    ioctl(fe_fd, FE_READ_SIGNAL_STRENGTH, &_signal);
    ioctl(fe_fd, FE_READ_SNR, &snr);
    ioctl(fe_fd, FE_READ_BER, &ber);
    ioctl(fe_fd, FE_READ_UNCORRECTED_BLOCKS, &uncorrected_blocks);

    printf("status %02x | signal %3u%% | snr %3u%% | ber %d | unc %d | ",
        status, (_signal * 100) / 0xffff, (snr * 100) / 0xffff, ber, uncorrected_blocks);
    if (status & FE_HAS_LOCK) {
        printf("FE_HAS_LOCK");
    }
    printf("\n");

    return (status & FE_HAS_LOCK) ? 1 : 0;
}


int check_frontend(int fe_fd)
{
    fe_status_t status;
    do {
        ioctl(fe_fd, FE_READ_STATUS, &status);
        print_frontend_stats(fe_fd);
        if (status & FE_HAS_LOCK) {
            break;
        }
        usleep(1000000);
    } while (!timeout_flag);
    print_frontend_stats(fe_fd);

    return (status & FE_HAS_LOCK) ? 1 : 0;
}
