// $CC -I/usr/include/libxml2 -I. -L ../../lib -l xml2 xml.c -o xml


#include <string.h>
#include <stdio.h>

#include "dvb.h"
#include "parse_dvb.h"
//#include "dump_dvb.h"
#include "xml.h"

//----------------------------------------------------------------------------
static char str2[0x60];

//!!!non-reentrant
static char *to_xml(char *str)
{
    int j;

    j = 0;
    while(*str) {
        if (*str == '&') {
            str2[j++] = '&';
            str2[j++] = 'a';
            str2[j++] = 'm';
            str2[j++] = 'p';
            str2[j++] = ';';
        }
        else
        if (*str == '<') {
            str2[j++] = '&';
            str2[j++] = 'l';
            str2[j++] = 't';
            str2[j++] = ';';
        }
        else
        if (*str == '>') {
            str2[j++] = '&';
            str2[j++] = 'g';
            str2[j++] = 't';
            str2[j++] = ';';
        }
        else {
            str2[j++] = *str;
        }
        str++;
    }
    str2[j++] = '\0';
    return str2;
}

//----------------------------------------------------------------------------
int init_transponder()
{
    int t,s;

    num_transponders = 0;
    for (t=0; t<10; t++) {
        transponder[t].network_id = -1;
        transponder[t].network_name[0] = '\0';
        transponder[t].frequency = 0;
        transponder[t].bandwidth = 0;
        transponder[t].fec_high = 0;
        transponder[t].fec_low = 0;
        transponder[t].modulation = 0;
        transponder[t].transmission_mode = 0;
        transponder[t].guard_interval = 0;
        transponder[t].hierarchy = 0;
        transponder[t].original_network_id = -1;
        transponder[t].transport_stream_id = -1;
        transponder[t].num_services = 0;
        for (s=0; s<0x20; s++) {
            transponder[t].service[s].transponder_id = -1;
            transponder[t].service[s].lcn = -1;
            transponder[t].service[s].service_id = -1;
            transponder[t].service[s].service_type = 0;
            transponder[t].service[s].provider_name[0] = '\0';
            transponder[t].service[s].service_name[0] = '\0';
            transponder[t].service[s].pmt_pid = -1;
            transponder[t].service[s].video_pid = -1;
            transponder[t].service[s].audio_pid = -1;
            transponder[t].service[s].audio2_pid = -1;
            transponder[t].service[s].pcr_pid = -1;
            transponder[t].service[s].subtitles_pid = -1;
            transponder[t].service[s].teletext_pid = -1;
            transponder[t].service[s].dsmcc_pid = -1;
            transponder[t].service[s].logo[0] = '\0';
        }
    }
    return 0;
}

//recursive
static int parse_xml(xmlDoc *doc, xmlNode *cur, int t, int s)
{
    xmlChar    *key;

    while (cur) {
//        printf("(%d,%d) name: %s\n", t,s, cur->name);

        if (strncasecmp(cur->name, "transponder", 11) == 0) {
            t++;
            num_transponders = t+1;
            s = -1;
            if (cur->children) parse_xml(doc, cur->children, t, s);
        }

        else
        if (strncasecmp(cur->name, "service\0", 8) == 0) {                //without /0, also matches "service_name" ...
            s++;
            transponder[t].num_services = s+1;
            if (cur->children) parse_xml(doc, cur->children, t, s);
        }

        else {
            if (!xmlIsBlankNode(cur)) {
                key = xmlNodeListGetString(doc, cur->children, 1);

                if (t == -1) {
                    printf("\noops\n");
                    exit(1);
                }
                else
                if (s == -1) {
                    if (strncasecmp(cur->name, "network_id\0",          11) == 0) { transponder[t].network_id          = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "network_name\0",        13) == 0) { if (key) strncpy(transponder[t].network_name, key, 0x40); }
                    if (strncasecmp(cur->name, "frequency\0",           10) == 0) { transponder[t].frequency           = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "bandwidth\0",           10) == 0) { transponder[t].bandwidth           = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "fec_high\0",             9) == 0) { transponder[t].fec_high            = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "fec_low\0",              8) == 0) { transponder[t].fec_low             = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "modulation\0",          11) == 0) { transponder[t].modulation          = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "transmission_mode\0",   18) == 0) { transponder[t].transmission_mode   = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "guard_interval\0",      15) == 0) { transponder[t].guard_interval      = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "hierarchy\0",           10) == 0) { transponder[t].hierarchy           = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "original_network_id\0", 20) == 0) { transponder[t].original_network_id = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "transport_stream_id\0", 20) == 0) { transponder[t].transport_stream_id = strtoul(key,NULL,16); }
                }
                else {
                    if (strncasecmp(cur->name, "transponder_id\0", 15) == 0) { transponder[t].service[s].transponder_id = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "lcn\0",             4) == 0) { transponder[t].service[s].lcn            = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "service_id\0",     11) == 0) { transponder[t].service[s].service_id     = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "service_type\0",   13) == 0) { transponder[t].service[s].service_type   = strtoul(key,NULL,10); }
                    if (strncasecmp(cur->name, "provider_name\0",  14) == 0) { if (key) strncpy(transponder[t].service[s].provider_name, key, 0x40); }
                    if (strncasecmp(cur->name, "service_name\0",   13) == 0) { if (key) strncpy(transponder[t].service[s].service_name,  key, 0x40); }
                    if (strncasecmp(cur->name, "pmt_pid\0",         8) == 0) { transponder[t].service[s].pmt_pid        = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "video_pid\0",      10) == 0) { transponder[t].service[s].video_pid      = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "audio_pid\0",      10) == 0) { transponder[t].service[s].audio_pid      = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "audio2_pid\0",     11) == 0) { transponder[t].service[s].audio2_pid     = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "pcr_pid\0",         8) == 0) { transponder[t].service[s].pcr_pid        = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "subtitles_pid\0",  14) == 0) { transponder[t].service[s].subtitles_pid  = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "teletext_pid\0",   13) == 0) { transponder[t].service[s].teletext_pid   = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "dsmcc_pid\0",      10) == 0) { transponder[t].service[s].dsmcc_pid      = strtoul(key,NULL,16); }
                    if (strncasecmp(cur->name, "logo\0",            5) == 0) { if (key) strncpy(transponder[t].service[s].logo, key, 0x40); }
                }

                xmlFree(key);
            }
        }
        cur = cur->next;
    }

    return 0;
}

int import_transponder(char *filename)
{
    xmlDoc     *doc = NULL;
    xmlNode    *root_element = NULL;

    doc = xmlReadFile(filename, NULL, 0);
    if (doc == NULL) {
        printf("error: could not parse file %s\n", filename);
        return -1;
    }
    root_element = xmlDocGetRootElement(doc);                   //channels
    if (strncasecmp(root_element->name, "channels", 8)) {
        printf("error: bad root element: %s\n", root_element->name);
        xmlFreeDoc(doc);
        return -1;
    }

    parse_xml(doc, root_element->children, -1, -1);

    xmlFreeDoc(doc);
    return 0;
}

void dump_transponder(char *filename)
{
    FILE *Op = 0;
    int t,s;

    if (filename) {
        Op = fopen(filename, "wt");
    }
    if (!Op) {
        Op = stdout;
    }
    for (t=0; t<num_transponders; t++) {
        fprintf(Op,"network_id          = %04x\n", transponder[t].network_id);
        fprintf(Op,"network_name        = %s\n", transponder[t].network_name);
        fprintf(Op,"frequency           = %d\n", transponder[t].frequency);
        fprintf(Op,"bandwidth           = %d (%s)\n", transponder[t].bandwidth, dvb_enumtype(bandwidth_str,transponder[t].bandwidth));
        fprintf(Op,"fec_high            = %d (%s)\n", transponder[t].fec_high, dvb_enumtype(code_rate_str,transponder[t].fec_high));
        fprintf(Op,"fec_low             = %d (%s)\n", transponder[t].fec_low, dvb_enumtype(code_rate_str,transponder[t].fec_low));
        fprintf(Op,"modulation          = %d (%s)\n", transponder[t].modulation, dvb_enumtype(constellation_str,transponder[t].modulation));
        fprintf(Op,"transmission_mode   = %d (%s)\n", transponder[t].transmission_mode, dvb_enumtype(transmission_mode_str,transponder[t].transmission_mode));
        fprintf(Op,"guard_interval      = %d (%s)\n", transponder[t].guard_interval, dvb_enumtype(guard_interval_str,transponder[t].guard_interval));
        fprintf(Op,"hierarchy           = %d (%s)\n", transponder[t].hierarchy, dvb_enumtype(hierarchy_str,transponder[t].hierarchy));
        fprintf(Op,"original_network_id = %04x\n", transponder[t].original_network_id);
        fprintf(Op,"transport_stream_id = %04x\n", transponder[t].transport_stream_id);
        fprintf(Op,"num_services  = %d\n", transponder[t].num_services);
        for (s=0; s<transponder[t].num_services; s++) {
        fprintf(Op,"  lcn               = %d\n", transponder[t].service[s].lcn);
        fprintf(Op,"  provider_name     = %s\n", transponder[t].service[s].provider_name);
        fprintf(Op,"  service_name      = %s\n", transponder[t].service[s].service_name);
        fprintf(Op,"  service_type      = %d (%s)\n", transponder[t].service[s].service_type, dvb_enumtype(service_type_str,transponder[t].service[s].service_type));
        fprintf(Op,"  service_id        = %04x\n", transponder[t].service[s].service_id);
        fprintf(Op,"  pmt_pid           = %04x\n", transponder[t].service[s].pmt_pid);
        fprintf(Op,"  video_pid         = %04x\n", transponder[t].service[s].video_pid);
        fprintf(Op,"  audio_pid         = %04x\n", transponder[t].service[s].audio_pid);
        fprintf(Op,"  pcr_pid           = %04x\n", transponder[t].service[s].pcr_pid);
        fprintf(Op,"  logo              = %s\n", transponder[t].service[s].logo);
        fprintf(Op,"\n");
        }
    }

    if (Op && (Op != stdout)) fclose(Op);
}

int export_transponder(char *filename)
{
    FILE *Op = 0;
    int t,s;

    if (filename) {
        Op = fopen(filename, "wt");
        if (!Op) return -1;
    }
    if (!Op) {
        Op = stdout;
    }

    fprintf(Op,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    fprintf(Op,"<?xml-stylesheet type=\"text/xsl\" href=\"channels.xsl\"?>\n");

    fprintf(Op,"<channels>\n");
    for (t=0; t<num_transponders; t++) {
        fprintf(Op,"  <transponder>\n");
        fprintf(Op,"    <network_id>0x%04x</network_id>\n", transponder[t].network_id);
        fprintf(Op,"    <network_name>%s</network_name>\n", to_xml(transponder[t].network_name));
        fprintf(Op,"    <frequency>%d</frequency>\n", transponder[t].frequency);
        fprintf(Op,"    <bandwidth>%d (%s)</bandwidth>\n", transponder[t].bandwidth, dvb_enumtype(bandwidth_str,transponder[t].bandwidth));
        fprintf(Op,"    <fec_high>%d (%s)</fec_high>\n", transponder[t].fec_high, dvb_enumtype(code_rate_str,transponder[t].fec_high));
        fprintf(Op,"    <fec_low>%d (%s)</fec_low>\n", transponder[t].fec_low, dvb_enumtype(code_rate_str,transponder[t].fec_low));
        fprintf(Op,"    <modulation>%d (%s)</modulation>\n", transponder[t].modulation, dvb_enumtype(constellation_str,transponder[t].modulation));
        fprintf(Op,"    <transmission_mode>%d (%s)</transmission_mode>\n", transponder[t].transmission_mode, dvb_enumtype(transmission_mode_str,transponder[t].transmission_mode));
        fprintf(Op,"    <guard_interval>%d (%s)</guard_interval>\n", transponder[t].guard_interval, dvb_enumtype(guard_interval_str,transponder[t].guard_interval));
        fprintf(Op,"    <hierarchy>%d (%s)</hierarchy>\n", transponder[t].hierarchy, dvb_enumtype(hierarchy_str,transponder[t].hierarchy));
        fprintf(Op,"    <original_network_id>0x%04x</original_network_id>\n", transponder[t].original_network_id);
        fprintf(Op,"    <transport_stream_id>0x%04x</transport_stream_id>\n", transponder[t].transport_stream_id);
        for (s=0; s<transponder[t].num_services; s++) {
            if (transponder[t].service[s].lcn > 0) {
                fprintf(Op,"    <service>\n");
                fprintf(Op,"      <lcn>%d</lcn>\n", transponder[t].service[s].lcn);
                fprintf(Op,"      <provider_name>%s</provider_name>\n", to_xml(transponder[t].service[s].provider_name));
                fprintf(Op,"      <service_name>%s</service_name>\n", to_xml(transponder[t].service[s].service_name));
                fprintf(Op,"      <service_type>%d (%s)</service_type>\n", transponder[t].service[s].service_type, dvb_enumtype(service_type_str,transponder[t].service[s].service_type));
                fprintf(Op,"      <service_id>0x%04x</service_id>\n", transponder[t].service[s].service_id);
                fprintf(Op,"      <pmt_pid>0x%04x</pmt_pid>\n", transponder[t].service[s].pmt_pid);
                fprintf(Op,"      <video_pid>0x%04x</video_pid>\n", transponder[t].service[s].video_pid);
                fprintf(Op,"      <audio_pid>0x%04x</audio_pid>\n", transponder[t].service[s].audio_pid);
                fprintf(Op,"      <pcr_pid>0x%04x</pcr_pid>\n", transponder[t].service[s].pcr_pid);
                fprintf(Op,"      <logo>logos/%s.png</logo>\n", to_xml(transponder[t].service[s].service_name));
                fprintf(Op,"    </service>\n");
            }
        }
        fprintf(Op,"  </transponder>\n");
    }
    fprintf(Op,"</channels>\n");
    fprintf(Op,"\n");

    if (Op && (Op != stdout)) fclose(Op);
    return 0;
}

void close_transponder()
{
    xmlCleanupParser();
}
